from compas.data import Data


class Value(Data):
    def __init__(self, value, value_type, options=None):
        super(Value, self).__init__()
        self._value = value
        self._value_type = value_type
        self._check_type(value)

        self._options = None
        if options is not None:
            self.options = options

    def __repr__(self):
        return "<{}: {}>".format(self.__class__.__name__, self.value)

    def _check_type(self, value):
        assert type(value) is self.value_type, "{} is not of type {}".format(value, self.value_type)

    def _check_options(self, value):
        if self.options is not None:
            assert value in self.options, "Value must be one of {}".format(self.options)

    def check(self, value):
        self._check_type(value)
        self._check_options(value)

    def cast(self, value):
        try:
            return self.value_type(value)
        except ValueError:
            raise ValueError("Cannot cast {} to {}".format(value, self.value_type))

    def set(self, value, raise_error=True, cast=True):
        try:
            if cast:
                value = self.cast(value)
            self.value = value
            return True
        except (AssertionError, ValueError) as e:
            if raise_error:
                raise e
            else:
                return e

    @property
    def value(self):
        return self._value

    @value.setter
    def value(self, value):
        self.check(value)
        self._value = value

    @property
    def value_type(self):
        return self._value_type

    @property
    def options(self):
        return self._options

    @options.setter
    def options(self, options):
        assert isinstance(options, list), "Options must be a list"
        for option in options:
            self._check_type(option)
        self._options = options

    @property
    def data(self):
        return {
            "value": self.value,
            "value_type": self.value_type.__name__,
            "options": self.options,
        }

    @data.setter
    def data(self, data):
        self._value = data["value"]
        self._value_type = eval(data["value_type"])  # TODO: support imported types
        self._options = data["options"]



import decimal
import traceback

import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore

from compas.colors import Color
from compas_rui.values import Settings
from compas_rui.values import Value


class CustomCell(Eto.Forms.CustomCell):
    def OnGetIdentifier(self, args):
        return str(args.Item.GetValue(0))

    def OnCreateCell(self, args):
        item = args.Item

        if item is None:
            return Eto.Forms.Label()

        value = item.GetValue(1)
        valueobj = item.GetValue(2)

        if isinstance(valueobj, Value):
            if valueobj.value_type is bool:
                control = Eto.Forms.CheckBox()
                control.Checked = value

                def on_checked(sender, e):
                    item.SetValue(1, control.Checked)

                control.CheckedChanged += on_checked

            elif valueobj.value_type is int:
                control = Eto.Forms.NumericUpDown()
                control.Value = value

                def on_value_changed(sender, e):
                    item.SetValue(1, int(control.Value))

                control.ValueChanged += on_value_changed

            elif valueobj.value_type is float:
                control = Eto.Forms.NumericUpDown()
                control.Value = value
                d = decimal.Decimal(str(value)).as_tuple()
                control.DecimalPlaces = -d.exponent
                control.MaximumDecimalPlaces = 3
                control.MinValue = 0
                control.Increment = 1e-3

                def on_value_changed(sender, e):
                    value = float(control.Value)
                    item.SetValue(1, value)

                control.ValueChanged += on_value_changed

            elif valueobj.value_type is str:
                control = Eto.Forms.TextBox()
                control.Text = value

                def on_text_changed(sender, e):
                    item.SetValue(1, str(control.Text))

                control.TextChanged += on_text_changed

            elif valueobj.value_type is Color:
                control = Eto.Forms.ColorPicker()
                control.Value = Eto.Drawing.Color.FromArgb(*value.rgb255)

                def on_value_changed(sender, e):
                    color = Eto.Drawing.Color(control.Value)
                    item.SetValue(1, Color(color.R, color.G, color.B))

                control.ValueChanged += on_value_changed

            else:
                control = Eto.Forms.Label(Text=str(valueobj))

        else:
            control = Eto.Forms.Label()

        control.Size = Eto.Drawing.Size(100, 25)

        return control


class SettingsForm(Eto.Forms.Dialog[bool]):
    def __init__(self, settings, title="Settings", width=500, height=500, use_tab=False):
        super().__init__()

        self._names = None
        self._values = None
        self.settings = settings
        self.use_tab = use_tab

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = True
        self.MinimumSize = Eto.Drawing.Size(0.5 * width, 0.5 * height)
        self.ClientSize = Eto.Drawing.Size(width, height)

        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical(Eto.Drawing.Padding(0, 0, 0, 0), Eto.Drawing.Size(0, 0), True, True)

        if self.use_tab:
            if not isinstance(settings, dict):
                raise ValueError("Settings must be a dictionary of compas_ui.values.Settings objects.")

            control = Eto.Forms.TabControl()
            control.TabPosition = Eto.Forms.DockPosition.Top

            self.tables = {}
            for key, settings in self.settings.items():
                if not isinstance(settings, Settings):
                    raise ValueError("Settings must be a dictionary of compas_ui.values.Settings objects.")

                tab = Eto.Forms.TabPage()
                tab.Text = key
                control.Pages.Add(tab)
                table = self.map_tree(settings)
                tab.Content = self.tables[key] = table
            layout.AddRow(control)

        else:
            if not isinstance(settings, Settings):
                raise ValueError("The settings must type compas_ui.values.Settings objects.")

            self.table = self.map_tree(settings)
            layout.AddRow(self.table)

        layout.EndVertical()
        layout.BeginVertical(Eto.Drawing.Padding(12, 18, 12, 24), Eto.Drawing.Size(6, 0), False, False)
        layout.AddRow(None, self.ok, self.cancel)
        layout.EndVertical()

        self.Content = layout

    def map_tree(self, settings):
        """Create the items for the form."""
        table = Eto.Forms.TreeGridView()
        table.ShowHeader = True

        column = Eto.Forms.GridColumn()
        column.HeaderText = "Key"
        column.Editable = False
        column.Sortable = True
        column.Expand = True
        column.DataCell = Eto.Forms.TextBoxCell(0)
        table.Columns.Add(column)

        column = Eto.Forms.GridColumn()
        column.HeaderText = "Value"
        column.Editable = True
        column.Sortable = False
        column.DataCell = Eto.Forms.TextBoxCell(0)
        column.DataCell = CustomCell()
        table.Columns.Add(column)

        def add_items(parent, items):
            keys = list(items.keys())
            keys.sort()
            for key in keys:
                value = items[key]
                if isinstance(value, Value):
                    item = Eto.Forms.TreeGridItem()
                    item.Values = (key, value.value, value)
                elif isinstance(value, dict):
                    item = Eto.Forms.TreeGridItem()
                    item.Values = (key, None, None)
                    add_items(item.Children, value)
                parent.Add(item)

        treecollection = Eto.Forms.TreeGridItemCollection()
        add_items(treecollection, settings.grouped_items)
        table.DataStore = treecollection
        return table

    @property
    def ok(self):
        self.DefaultButton = Eto.Forms.Button()
        self.DefaultButton.Text = "OK"
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    @property
    def cancel(self):
        self.AbortButton = Eto.Forms.Button()
        self.AbortButton.Text = "Cancel"
        self.AbortButton.Click += self.on_cancel
        return self.AbortButton

    @property
    def names(self):
        return self._names

    @property
    def values(self):
        return self._values

    def on_ok(self, sender, event):
        try:

            def set_value(items, setting):
                for item in items:
                    key = item.GetValue(0)
                    value = item.GetValue(1)
                    valueobj = item.GetValue(2)
                    if valueobj:
                        valueobj.set(value)
                    else:
                        set_value(item.Children, setting[key])

            if not self.use_tab:
                set_value(self.table.DataStore.Items, self.settings.grouped_items)
            else:
                for key in self.settings:
                    set_value(self.tables[key].DataStore.Items, self.settings[key].grouped_items)

        except Exception as e:
            traceback.print_exc()
            print("ERROR:", e)
            self.Close(False)

        print("OK")
        self.Close(True)

    def on_cancel(self, sender, event):
        self.Close(False)

    def show(self):
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)


if __name__ == "__main__":
    from compas_rui.values import BoolValue
    from compas_rui.values import ColorValue
    from compas_rui.values import Settings

    settings = {
        "Session": Settings({"session.A": BoolValue(True), "tna.B": ColorValue(Color.red())}),
    }

    form = SettingsForm(settings, use_tab=True)
    form.show()



import ast

import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore


class NamedValuesForm(Eto.Forms.Dialog[bool]):
    """Eto form for displaying and editing named values.

    Note that when providing a new value for a name-value pair,
    only literal Python structures are allowed: strings, bytes, numbers, tuples, lists, dicts, sets, booleans, and None
    Other types of values will be ignored.

    Parameters
    ----------
    names : list[str]
        The names of the values.
    values: list[str, byte, float, int, tuple, list, dict, set, bool, None]
        The values.
        Only literal Python structures are allowed.
        Other types of values are ignored.
    title : str, optional
        The title of the form window.
    width : int, optional
        The width of the form window.
    height : int, optional
        The height of the form window.

    """

    def __init__(
        self,
        names,  # type: list[str]
        values,  # type: list
        title="Named Values",  # type: str
        width=500,  # type: int
        height=500,  # type: int
    ):
        # type: (...) -> None

        super().__init__()

        def on_cell_formatting(sender, e):
            try:
                if not e.Column.Editable:
                    e.ForegroundColor = Eto.Drawing.Colors.DarkGray
            except Exception as e:
                print(e)

        self.attributes = dict(zip(names, values))
        self.names = names
        self.values = [str(value) for value in values]

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = True
        self.MinimumSize = Eto.Drawing.Size(0.5 * width, 0.5 * height)
        self.ClientSize = Eto.Drawing.Size(width, height)

        self.table = table = Eto.Forms.GridView()
        table.ShowHeader = True

        column = Eto.Forms.GridColumn()
        column.HeaderText = "Name"
        column.Editable = False
        column.Expand = True
        column.DataCell = Eto.Forms.TextBoxCell(0)
        table.Columns.Add(column)

        column = Eto.Forms.GridColumn()
        column.HeaderText = "Value"
        column.Editable = True
        column.DataCell = Eto.Forms.TextBoxCell(1)
        table.Columns.Add(column)

        collection = []
        for name, value in zip(self.names, self.values):
            item = Eto.Forms.GridItem()
            item.Values = (name, value)
            collection.append(item)
        table.DataStore = collection

        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical(Eto.Drawing.Padding(0, 0, 0, 0), Eto.Drawing.Size(0, 0), True, True)
        layout.AddRow(table)
        layout.EndVertical()
        layout.BeginVertical(Eto.Drawing.Padding(12, 18, 12, 24), Eto.Drawing.Size(6, 0), False, False)
        layout.AddRow(None, self.ok, self.cancel)
        layout.EndVertical()

        self.Content = layout

    @property
    def ok(self):
        self.DefaultButton = Eto.Forms.Button()
        self.DefaultButton.Text = "OK"
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    @property
    def cancel(self):
        self.AbortButton = Eto.Forms.Button()
        self.AbortButton.Text = "Cancel"
        self.AbortButton.Click += self.on_cancel
        return self.AbortButton

    def on_ok(self, sender, event):
        try:
            for row in self.table.DataStore:
                name = row.GetValue(0)
                value = row.GetValue(1)
                if value != "-":
                    try:
                        value = ast.literal_eval(value)
                    except Exception as e:
                        print(e)
                self.attributes[name] = value
        except Exception as e:
            print(e)
            self.Close(False)
        self.Close(True)

    def on_cancel(self, sender, event):
        self.Close(False)

    def show(self):
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)


if __name__ == "__main__":
    form = NamedValuesForm(names=["a", "b"], values=[1, 2])
    if form.show():
        print(form.attributes)



import compas_rhino


class CommandMenu(object):
    def __init__(self, menu):
        self.menu = menu

    def select_action(self):
        def _select(message, options):
            if not options:
                return
            names = [option["name"] for option in options]
            name = compas_rhino.rs.GetString(message, names[0], names)
            if not name:
                return
            if name not in names:
                raise Exception("This option is not valid: {}".format(name))
            for option in options:
                if option["name"] == name:
                    break
            if "action" in option:
                return option

            message = option["message"]
            options = option.get("options")

            return _select(message, options)

        return _select(self.menu["message"], self.menu["options"])


class CommandAction(object):
    def __init__(self, name, action):
        self.name = name
        self.action = action

    def __call__(self, *args, **kwargs):
        return self.action(*args, **kwargs)



from compas.plugins import plugin
from compas.scene.context import register

from compas.datastructures import Mesh
from .meshobject import RUIMeshObject


@plugin(category="factories", pluggable_name="register_scene_objects", requires=["Rhino"])
def register_scene_objects_rhino():
    register(Mesh, RUIMeshObject, context="Rhino")



from __future__ import print_function

import os


__author__ = ["tom van mele"]
__copyright__ = "ETH Zurich - Block Research Group"
__license__ = "MIT License"
__email__ = "tom.v.mele@gmail.com"
__version__ = "0.4.1"


HERE = os.path.dirname(__file__)

HOME = os.path.abspath(os.path.join(HERE, "../../"))
DATA = os.path.abspath(os.path.join(HOME, "data"))
DOCS = os.path.abspath(os.path.join(HOME, "docs"))
TEMP = os.path.abspath(os.path.join(HOME, "temp"))


__all__ = ["HOME", "DATA", "DOCS", "TEMP"]

__all_plugins__ = [
    "compas_rui.scene",
]



from typing import Optional

import Rhino  # type: ignore
import rhinoscriptsyntax as rs  # type: ignore

import compas_rhino.conversions
import compas_rhino.objects
from compas.datastructures import Mesh
from compas.geometry import Point
from compas.itertools import flatten
from compas_rhino.scene import RhinoMeshObject
from compas_rui.forms import NamedValuesForm


class RUIMeshObject(RhinoMeshObject):
    mesh: Mesh

    # =============================================================================
    # =============================================================================
    # =============================================================================
    # Select
    # =============================================================================
    # =============================================================================
    # =============================================================================

    def select_vertices(self, message="Select Vertices") -> list[int]:
        options = ["All", "Boundary", "Degree", "EdgeLoop", "EdgeStrip", "Manual"]
        option = rs.GetString(message=message, strings=options)
        if not option:
            return

        if option == "All":
            vertices = self.select_vertices_all()

        elif option == "Boundary":
            vertices = self.select_vertices_boundary()

        elif option == "Degree":
            vertices = self.select_vertices_degree()

        elif option == "EdgeLoop":
            vertices = self.select_vertices_edgeloop()

        elif option == "EdgeStrip":
            vertices = self.select_vertices_edgestrip()

        elif option == "Manual":
            vertices = self.select_vertices_manual()

        vertex_guid = {vertex: guid for guid, vertex in self._guid_vertex.items()}
        guids = [vertex_guid[vertex] for vertex in vertices]

        rs.UnselectAllObjects()
        rs.SelectObjects(guids)

        return vertices

    def select_vertices_all(self):
        return list(self.mesh.vertices())

    def select_vertices_boundary(self):
        return list(set(flatten(self.mesh.vertices_on_boundaries())))

    def select_vertices_degree(self):
        D = rs.GetInteger(message="Vertex Degree", number=2, minimum=1)
        D = D or 0
        vertices = list(self.mesh.vertices_where(vertex_degree=D))
        return vertices

    def select_vertices_edgeloop(self):
        guids = compas_rhino.objects.select_lines(message="Select Loop Edges")
        edges = [self._guid_edge[guid] for guid in guids if guid in self._guid_edge] if guids else []
        temp = []
        for edge in edges:
            for u, v in self.mesh.edge_loop(edge):
                temp.append(u)
                temp.append(v)
        vertices = list(set(temp))
        return vertices

    def select_vertices_edgestrip(self):
        raise NotImplementedError

    def select_vertices_manual(self):
        guids = compas_rhino.objects.select_points(message="Select Vertices")
        vertices = [self._guid_vertex[guid] for guid in guids if guid in self._guid_vertex] if guids else []
        return vertices

    def select_edges(self, message="Select Edges") -> list[tuple[int, int]]:
        options = ["All", "Boundary", "EdgeLoop", "EdgeStrip", "Manual"]
        option = rs.GetString(message=message, strings=options)
        if not option:
            return

        if option == "All":
            edges = self.select_edges_all()

        elif option == "Boundary":
            edges = self.select_edges_boundary()

        elif option == "EdgeLoop":
            edges = self.select_edges_loop()

        elif option == "EdgeStrip":
            edges = self.select_edges_strip()

        elif option == "Manual":
            edges = self.select_edges_manual()

        edges = [(u, v) if self.mesh.has_edge((u, v)) else (v, u) for u, v in edges]
        edge_guid = {edge: guid for guid, edge in self._guid_edge.items()}
        guids = [edge_guid[edge] for edge in edges]

        rs.UnselectAllObjects()
        rs.SelectObjects(guids)

        return edges

    def select_edges_all(self):
        return list(self.mesh.edges())

    def select_edges_boundary(self):
        return list(set(flatten(self.mesh.edges_on_boundaries())))

    def select_edges_loop(self):
        guids = compas_rhino.objects.select_lines(message="Select Loop Edges")
        edges = []
        for guid in guids:
            edge = self._guid_edge[guid]
            for edge in self.mesh.edge_loop(edge):
                edges.append(edge)
        return edges

    def select_edges_strip(self):
        guids = compas_rhino.objects.select_lines(message="Select Strip Edges")
        edges = []
        for guid in guids:
            edge = self._guid_edge[guid]
            for edge in self.mesh.edge_strip(edge):
                edges.append(edge)
        return edges

    def select_edges_manual(self):
        guids = compas_rhino.objects.select_lines(message="Select Edges")
        edges = [self._guid_edge[guid] for guid in guids if guid in self._guid_edge] if guids else []
        return edges

    def select_faces(self, message="Select Faces") -> list[int]:
        options = ["All", "Boundary", "Strip", "Manual"]
        option = rs.GetString(message=message, strings=options)
        if not option:
            return

        if option == "All":
            faces = self.select_faces_all()

        elif option == "Boundary":
            faces = self.select_faces_boundary()

        elif option == "Strip":
            faces = self.select_faces_strip()

        elif option == "Manual":
            faces = self.select_faces_manual()

        face_guid = {face: guid for guid, face in self._guid_face.items()}
        guids = [face_guid[face] for face in faces]

        rs.UnselectAllObjects()
        rs.SelectObjects(guids)

        return faces

    def select_faces_all(self):
        return list(self.mesh.faces())

    def select_faces_boundary(self):
        return list(set(flatten(self.mesh.faces_on_boundaries())))

    def select_faces_strip(self):
        raise NotImplementedError

    def select_faces_manual(self):
        guids = compas_rhino.objects.select_meshes(message="Select Faces")
        faces = [self._guid_face[guid] for guid in guids if guid in self._guid_face] if guids else []
        return faces

    # =============================================================================
    # =============================================================================
    # =============================================================================
    # Draw
    # =============================================================================
    # =============================================================================
    # =============================================================================

    # =============================================================================
    # =============================================================================
    # =============================================================================
    # Modify
    # =============================================================================
    # =============================================================================
    # =============================================================================

    def update_attributes(self) -> bool:
        names = sorted(self.mesh.attributes.keys())
        values = [str(self.mesh.attributes[name]) for name in names]

        form = NamedValuesForm(names=names, values=values)
        if form.show():
            self.mesh.attributes.update(form.attributes)
            return True
        return False

    def update_vertex_attributes(self, vertices: list[int], names: Optional[list[str]] = None) -> bool:
        if not vertices:
            return False

        names = names or sorted(self.mesh.default_vertex_attributes.keys())
        names = sorted([name for name in names if not name.startswith("_")])

        values = self.mesh.vertex_attributes(vertices[0], names)
        if len(vertices) > 1:
            for i, name in enumerate(names):
                for vertex in vertices[1:]:
                    if values[i] != self.mesh.vertex_attribute(vertex, name):
                        values[i] = "-"
                        break
        values = list(map(str, values))

        form = NamedValuesForm(names=names, values=values)
        if form.show():
            for name, value in form.attributes.items():
                if value == "-":
                    continue
                self.mesh.vertices_attribute(name=name, value=value, keys=vertices)
            return True
        return False

    def update_face_attributes(self, faces: list[int], names: Optional[list[str]] = None) -> bool:
        if not faces:
            return False

        names = names or sorted(self.mesh.default_face_attributes.keys())
        names = sorted([name for name in names if not name.startswith("_")])

        values = self.mesh.face_attributes(faces[0], names)
        if len(faces) > 1:
            for i, name in enumerate(names):
                for face in faces[1:]:
                    if values[i] != self.mesh.face_attribute(face, name):
                        values[i] = "-"
                        break
        values = list(map(str, values))

        form = NamedValuesForm(names=names, values=values)
        if form.show():
            for name, value in form.attributes.items():
                if value == "-":
                    continue
                self.mesh.faces_attribute(name=name, value=value, keys=faces)
            return True
        return False

    def update_edge_attributes(self, edges: list[tuple[int, int]], names: Optional[list[str]] = None) -> bool:
        if not edges:
            return False

        names = names or sorted(self.mesh.default_edge_attributes.keys())
        names = sorted([name for name in names if not name.startswith("_")])

        values = self.mesh.edge_attributes(edges[0], names)
        if len(edges) > 1:
            for i, name in enumerate(names):
                for edge in edges[1:]:
                    if values[i] != self.mesh.edge_attribute(edge, name):
                        values[i] = "-"
                        break
        values = list(map(str, values))

        form = NamedValuesForm(names=names, values=values)
        if form.show():
            for name, value in form.attributes.items():
                if value == "-":
                    continue
                self.mesh.edges_attribute(name=name, value=value, keys=edges)
            return True
        return False

    # =============================================================================
    # =============================================================================
    # =============================================================================
    # Move
    # =============================================================================
    # =============================================================================
    # =============================================================================

    def move(self) -> bool:
        color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor

        vertex_p0 = {v: Rhino.Geometry.Point3d(*self.mesh.vertex_coordinates(v)) for v in self.mesh.vertices()}
        vertex_p1 = {v: Rhino.Geometry.Point3d(*self.mesh.vertex_coordinates(v)) for v in self.mesh.vertices()}

        edges = list(self.mesh.edges())

        def OnDynamicDraw(sender, e):
            current = e.CurrentPoint
            vector = current - start
            for vertex in vertex_p1:
                vertex_p1[vertex] = vertex_p0[vertex] + vector
            for u, v in iter(edges):
                sp = vertex[u]
                ep = vertex[v]
                e.Display.DrawDottedLine(sp, ep, color)

        gp = Rhino.Input.Custom.GetPoint()

        gp.SetCommandPrompt("Point to move from?")
        gp.Get()

        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        start = gp.Point()

        gp = Rhino.Input.Custom.GetPoint()
        gp.SetCommandPrompt("Point to move to?")
        gp.DynamicDraw += OnDynamicDraw
        gp.Get()

        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        end = gp.Point()
        vector = compas_rhino.conversions.vector_to_compas(end - start)

        for _, attr in self.mesh.vertices(True):
            attr["x"] += vector[0]
            attr["y"] += vector[1]
            attr["z"] += vector[2]

        return True

    def move_vertex(
        self,
        vertex: int,
        constraint: Rhino.Geometry.GeometryBase = None,
        allow_off: bool = True,
    ) -> bool:
        def OnDynamicDraw(sender, e):
            for ep in nbrs:
                sp = e.CurrentPoint
                e.Display.DrawDottedLine(sp, ep, color)

        color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor
        nbrs = [self.mesh.vertex_coordinates(nbr) for nbr in self.mesh.vertex_neighbors(vertex)]
        nbrs = [Rhino.Geometry.Point3d(*xyz) for xyz in nbrs]

        gp = Rhino.Input.Custom.GetPoint()

        gp.SetCommandPrompt("Point to move to?")
        gp.DynamicDraw += OnDynamicDraw
        if constraint:
            gp.Constrain(constraint, allow_off)

        gp.Get()
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        self.mesh.vertex_attributes(vertex, "xyz", list(gp.Point()))
        return True

    def move_vertices(self, vertices: list[int]) -> bool:
        def OnDynamicDraw(sender, e):
            end = e.CurrentPoint
            vector = end - start
            for a, b in lines:
                a = a + vector
                b = b + vector
                e.Display.DrawDottedLine(a, b, color)
            for a, b in connectors:
                a = a + vector
                e.Display.DrawDottedLine(a, b, color)

        color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor
        lines = []
        connectors = []

        for vertex in vertices:
            a = self.mesh.vertex_coordinates(vertex)
            nbrs = self.mesh.vertex_neighbors(vertex)
            for nbr in nbrs:
                b = self.mesh.vertex_coordinates(nbr)
                line = [Rhino.Geometry.Point3d(*a), Rhino.Geometry.Point3d(*b)]
                if nbr in vertices:
                    lines.append(line)
                else:
                    connectors.append(line)

        gp = Rhino.Input.Custom.GetPoint()

        gp.SetCommandPrompt("Point to move from?")
        gp.Get()
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        start = gp.Point()

        gp.SetCommandPrompt("Point to move to?")
        gp.SetBasePoint(start, False)
        gp.DrawLineFromPoint(start, True)
        gp.DynamicDraw += OnDynamicDraw
        gp.Get()
        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        end = gp.Point()
        vector = compas_rhino.conversions.vector_to_compas(end - start)

        for vertex in vertices:
            point = Point(*self.mesh.vertex_attributes(vertex, "xyz"))
            self.mesh.vertex_attributes(vertex, "xyz", point + vector)
        return True

    def move_vertices_direction(self, vertices: list[int], direction: str) -> bool:
        def OnDynamicDraw(sender, e):
            draw = e.Display.DrawDottedLine
            end = e.CurrentPoint
            vector = end - start
            for a, b in lines:
                a = a + vector
                b = b + vector
                draw(a, b, color)
            for a, b in connectors:
                a = a + vector
                draw(a, b, color)

        direction = direction.lower()
        color = Rhino.ApplicationSettings.AppearanceSettings.FeedbackColor
        lines = []
        connectors = []

        for vertex in vertices:
            a = Rhino.Geometry.Point3d(*self.mesh.vertex_coordinates(vertex))
            nbrs = self.mesh.vertex_neighbors(vertex)
            for nbr in nbrs:
                b = Rhino.Geometry.Point3d(*self.mesh.vertex_coordinates(nbr))
                if nbr in vertices:
                    lines.append((a, b))
                else:
                    connectors.append((a, b))

        gp = Rhino.Input.Custom.GetPoint()
        gp.SetCommandPrompt("Point to move from?")
        gp.Get()

        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        start = gp.Point()

        if direction == "x":
            geometry = Rhino.Geometry.Line(start, start + Rhino.Geometry.Vector3d(1, 0, 0))
        elif direction == "y":
            geometry = Rhino.Geometry.Line(start, start + Rhino.Geometry.Vector3d(0, 1, 0))
        elif direction == "z":
            geometry = Rhino.Geometry.Line(start, start + Rhino.Geometry.Vector3d(0, 0, 1))
        elif direction == "xy":
            geometry = Rhino.Geometry.Plane(start, Rhino.Geometry.Vector3d(0, 0, 1))
        elif direction == "yz":
            geometry = Rhino.Geometry.Plane(start, Rhino.Geometry.Vector3d(1, 0, 0))
        elif direction == "zx":
            geometry = Rhino.Geometry.Plane(start, Rhino.Geometry.Vector3d(0, 1, 0))

        gp.SetCommandPrompt("Point to move to?")
        gp.SetBasePoint(start, False)
        gp.DrawLineFromPoint(start, True)
        gp.DynamicDraw += OnDynamicDraw

        if direction in ("x", "y", "z"):
            gp.Constrain(geometry)
        else:
            gp.Constrain(geometry, False)

        gp.Get()

        if gp.CommandResult() != Rhino.Commands.Result.Success:
            return False

        end = gp.Point()
        vector = compas_rhino.conversions.vector_to_compas(end - start)

        for vertex in vertices:
            point = self.mesh.vertex_point(vertex)
            self.mesh.vertex_attributes(vertex, "xyz", point + vector)

        return True

    # =============================================================================
    # =============================================================================
    # =============================================================================
    # Conduits
    # =============================================================================
    # =============================================================================
    # =============================================================================
    # =============================================================================

    def clear_conduits(self):
        pass

    def clear(self):
        super().clear()
        self.clear_conduits()



from .value import Value
from .boolvalue import BoolValue
from .numbervalue import IntValue
from .numbervalue import FloatValue
from .strvalue import StrValue
from .listvalue import ListValue
from .dictvalue import DictValue
from .colorvalue import ColorValue
from .settings import Settings

__all__ = [
    "Value",
    "BoolValue",
    "IntValue",
    "FloatValue",
    "StrValue",
    "ListValue",
    "DictValue",
    "ColorValue",
    "Settings",
]



from .about import AboutForm
from .namedvalues import NamedValuesForm
from .error import ErrorForm, error
from .filesystem import FileForm
from .filesystem import FolderForm
from .info import InfoForm
from .meshdata import MeshDataForm
from .settings import SettingsForm
from .scene import SceneObjectsForm
from .splash import SplashForm


__all__ = [
    "AboutForm",
    "NamedValuesForm",
    "CondaEnvsForm",
    "ErrorForm",
    "error",
    "FileForm",
    "FolderForm",
    "InfoForm",
    "MeshDataForm",
    "SearchPathsForm",
    "SettingsForm",
    "SceneObjectsForm",
    "SplashForm",
    "ToolbarForm",
]



import Eto.Forms  # type: ignore
import Rhino.UI  # type: ignore
import System  # type: ignore


class AboutForm:
    def __init__(
        self,
        title,  # type: str
        description,  # type: str
        version,  # type: str
        website,  # type: str
        copyright,  # type: str
        license,  # type: str
        designers=None,  # type: list[str] | None
        developers=None,  # type: list[str] | None
        documenters=None,  # type: list[str] | None
    ):
        # type: (...) -> None
        designers = designers or []
        developers = developers or []
        documenters = documenters or []

        self.dialog = Eto.Forms.AboutDialog()
        self.dialog.Copyright = copyright
        self.dialog.Designers = System.Array[System.String](designers)
        self.dialog.Developers = System.Array[System.String](developers)
        self.dialog.Documenters = System.Array[System.String](documenters)
        self.dialog.License = license
        self.dialog.ProgramDescription = description
        self.dialog.ProgramName = title
        self.dialog.Title = title
        self.dialog.Version = version
        self.dialog.Website = System.Uri(website)

    def show(self):
        self.dialog.ShowDialog(Rhino.UI.RhinoEtoApp.MainWindow)



import sys
import traceback
from functools import wraps

import Eto.Drawing  # type: ignore
import Eto.Forms  # type: ignore
import Rhino  # type: ignore
import Rhino.UI  # type: ignore


def error(title="Error", showLocalTraceback=True):
    def outer(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            try:
                return func(*args, **kwargs)
            except Exception as error:
                if showLocalTraceback:
                    exc_type, exc_value, exc_tb = sys.exc_info()
                    text = traceback.format_exception(exc_type, exc_value, exc_tb)
                    text = "".join(text)
                else:
                    text = str(error)
                form = ErrorForm(text, title=title)
                form.show()

        return wrapper

    return outer


ISSUE_TEMPLATE = """---
name: Bug report
about: Create a report to help us improve

---
<!-- The link below shows a list of known-issues and how to fix them -->
<!-- https://blockresearchgroup.gitbook.io/FF/documentation/known-issues -->
<!-- If the error you encounter is not in the list, please describe it as following -->
<!-- We thank you on the feedback -->

**Describe the bug**
```bash
%s
```

**To Reproduce**
Steps to reproduce the behavior:
1. Please provide an input 3dm file
2. Describe the command that causes the error
3. A screenshot of error behavior

**Expected behavior**
A clear and concise description of what you expected to happen.

**Screenshots**
If applicable, add screenshots or video record to help explain your problem.

**Desktop (please complete the following information):**
 - OS: %s
 - Rhino version %s
 - FoFin version %s

**Additional context**
Add any other context about the problem here.
"""


class ErrorForm(Eto.Forms.Dialog[bool]):
    def __init__(self, error="Unknown", title="Error", width=800, height=500):
        super().__init__()

        self.Title = title
        self.Padding = Eto.Drawing.Padding(0)
        self.Resizable = True
        self.MinimumSize = Eto.Drawing.Size(0.5 * width, 0.5 * height)
        self.ClientSize = Eto.Drawing.Size(width, height)

        textarea = Eto.Forms.TextArea()
        textarea.Text = error
        textarea.ReadOnly = True

        layout = Eto.Forms.DynamicLayout()
        layout.BeginVertical(Eto.Drawing.Padding(12, 12, 12, 0), Eto.Drawing.Size(0, 0), True, True)
        layout.AddRow(textarea)
        layout.EndVertical()
        layout.BeginVertical(Eto.Drawing.Padding(12, 12, 12, 18), Eto.Drawing.Size(6, 0), False, False)
        layout.AddRow(None, self.ok)
        layout.EndVertical()
        self.Content = layout

    @property
    def ok(self):
        self.DefaultButton = Eto.Forms.Button()
        self.DefaultButton.Text = "OK"
        self.DefaultButton.Click += self.on_ok
        return self.DefaultButton

    def on_ok(self, sender, event):
        self.Close(True)

    def show(self):
        return self.ShowModal(Rhino.UI.RhinoEtoApp.MainWindow)



from .dictvalue import DictValue
from .value import Value


class Settings(DictValue):
    def __init__(self, settings=None):
        if settings is None:
            settings = {}
        super(Settings, self).__init__(settings, Value)

    def __getitem__(self, key):
        return self.value[key].value

    def __setitem__(self, key, value):
        self.value[key].value = value

    def items(self):
        return self.value.items()

    def keys(self):
        return self.value.keys()

    def get(self, key, valueobject=False):
        if valueobject:
            return self.value[key]
        else:
            return self.value[key].value

    def update(self, settings):
        for k, v in settings.items():
            self.value[k] = v

    @property
    def grouped_items(self):
        def group(settings):
            groups = {}
            for key, value in settings.items():
                parts = key.split(".")
                if len(parts) > 1:
                    if parts[0] not in groups:
                        groups[parts[0]] = {}
                    subkey = ".".join(parts[1:])
                    groups[parts[0]][subkey] = value
                else:
                    groups[key] = value

            for key, value in groups.items():
                if isinstance(value, dict):
                    groups[key] = group(value)

            return groups

        return group(self.value)



from .value import Value


class DictValue(Value):
    def __init__(self, value=None, dict_value_type=None):
        if value is None:
            value = {}
        super(DictValue, self).__init__(value, dict)
        self._dict_value_type = dict_value_type
        self._check_dict_value_type(value)

    def _check_dict_value_type(self, value):
        for k, v in value.items():
            if not isinstance(k, str):
                raise ValueError("Dict key {} is not of type {}".format(k, str))

            if not isinstance(v, self.dict_value_type):
                raise ValueError("Dict value {}:{} is not of type {}".format(k, v, self.dict_value_type))

    def check(self, value):
        super(DictValue, self).check(value)
        self._check_dict_value_type(value)

    def __getitem__(self, key):
        return self.value[key]

    def __setitem__(self, key, value):
        if not isinstance(value, self.dict_value_type):
            raise ValueError("New value {} is not of type {}".format(value, self.dict_value_type))

        self.value[key] = value

    @property
    def dict_value_type(self):
        return self._dict_value_type

    @property
    def data(self):
        return {
            "value": self.value,
            "value_type": "dict",
            "dict_value_type": self.dict_value_type.__name__,
        }

    @data.setter
    def data(self, data):
        self._value = data["value"]
        self._dict_value_type = eval(data["dict_value_type"])
